#
#   Imandra Inc.
#
#   codelogician/modeling/metamodel.py
#

import logging
from collections import defaultdict
from pathlib import Path

from imandra.u.agents.code_logician.base import FormalizationStatus
from pydantic import BaseModel, Field, ValidationInfo, model_validator

from ..server.events import FileSystemEvent, FileSystemEventType
from ..sketch.sketch import Sketch
from ..strategy.cl_agent_state import CLAgentState, CLResult
from .dep_context import ModelDepContext
from .fs_sync import model_filesystem_sync
from .model import Model
from .task import ModelTask, TaskKind
from .task_creator import ModelTaskCreator

log = logging.getLogger(__name__)


class CycleSummary(BaseModel):
    rel_paths: list[str]
    edge_paths: list[tuple[str, str]]


class MetaModel(BaseModel):
    """
    Container for the set of models representing the whole project
    """

    src_dir_abs_path: str
    models: dict[str, Model]
    language: str = 'Python'

    dep_cycles: list[CycleSummary] = Field(default_factory=list)

    @model_validator(mode='after')
    def link_model_deps(self, info: ValidationInfo) -> 'MetaModel':
        # Only rebuild edges when explicitly asked (e.g., load-from-disk path).
        if not (info.context or {}).get('relink_from_str_edges', False):
            return self

        # Clear in-memory edges
        for m in self.models.values():
            m.dependencies.clear()
            m.rev_dependencies.clear()

        # Rebuild from persisted strings (deps first)
        for m in self.models.values():
            for dep_path in m.str_dependencies:
                dep = self.models.get(dep_path)
                if dep is not None:
                    m.dependencies.append(dep)

        # Rebuild rev-deps as inverse to guarantee consistency
        for m in self.models.values():
            for dep in m.dependencies:
                dep.rev_dependencies.append(m)

        return self

    def get_paths_with_dirs(self):
        """
        Return a list of paths with directory information
        """

        FILE_MARKER = '<files>'

        def attach(branch, trunk, model):
            parts = branch.split('/', 1)
            if len(parts) == 1:  # branch is a file
                trunk[FILE_MARKER].append(model)
            else:
                node, others = parts
                if node not in trunk:
                    trunk[node] = defaultdict(dict, ((FILE_MARKER, []),))
                attach(others, trunk[node], model)

        main_dict = defaultdict(dict, ((FILE_MARKER, []),))

        for path in sorted(self.models.keys()):
            attach(path, main_dict, self.models[path])

        return main_dict

    def list_models(self) -> list[Model]:
        """
        Return the list of all models
        """
        return list(self.models.values())

    def vgs(self):
        """
        Return all verification goals for all the models
        """
        return {m.rel_path: m.verification_goals() for m in self.models.values()}

    def decomps(self):
        """
        Return decomposition requests for all the models
        """
        return {m.rel_path: m.decomps() for m in self.models.values()}

    def opaques(self):
        """
        Return all the opaque functions for all the models
        """
        return {m.rel_path: m.opaque_funcs() for m in self.models.values()}

    def apply_cl_result(self, cl_result: CLResult) -> None:
        """
        Applies the result of running CodeLogician on a model
        """

        if cl_result.task.rel_path in self.models:
            model = self.models[cl_result.task.rel_path]

            # Rememeber that we may have multiple tasks outstanding simultaneously
            # so, we always keep the last one that we use
            if model.outstanding_task_ID == cl_result.task.task_id:
                dependencies = cl_result.task.dependencies
                model.apply_agent_state(cl_result.agent_state, dependencies)
                model.outstanding_task_ID = None
                model.outstanding_task_kind = None

                if cl_result.task.task_kind == TaskKind.ANALYZE:
                    model.analysis_vgs_attempted |= bool(cl_result.task.gen_vgs)
                    model.analysis_decomps_attempted |= bool(cl_result.task.gen_decomps)

                elif cl_result.task.task_kind == TaskKind.USER_IML_UPDATE:
                    # User edits implicitly “finalize” analysis for this baseline
                    model.analysis_vgs_attempted = True
                    model.analysis_decomps_attempted = True

                elif cl_result.task.task_kind == TaskKind.FORMALIZE:
                    # New baseline → allow analyze again
                    model.analysis_vgs_attempted = False
                    model.analysis_decomps_attempted = False

                log.info('CLResult applied to model successfully!')
            else:
                log.warning(
                    f'Model with path={cl_result.task.rel_path} has a newer Task ID [{model.outstanding_task_ID}], '
                    f' so ignoring this one [{cl_result.task.task_kind}:{cl_result.task.task_id}]!'
                )
        else:
            log.error(
                f'Attempting to apply CL Result to a non-existent model path: {cl_result.task.rel_path}!'
            )

    def update_model(self, rel_path: str, state: CLAgentState) -> None:
        """
        Update the model with a new agent state (formalization state form CL agent)
        """

        if rel_path in self.models:
            self.models[rel_path].apply_agent_state(state)
            log.info(
                f'AgentState successfully updated for model with path=[{rel_path}]'
            )
        else:
            log.error(
                f'Attempting to set agent state to a non-existent model path {rel_path}'
            )

    def get_next_tasks(self, user_wait_time: int | None = None) -> list[ModelTask]:
        log.debug('About to get_next_tasks')

        # Hard gate: do not schedule tasks if cycles exist
        if getattr(self, 'dep_cycles', []):
            log.warning(
                'Not scheduling tasks because dependency cycles are present: %s',
                [c.rel_paths for c in self.dep_cycles],
            )
            return []

        tasks: list[ModelTask] = []

        task_creator = ModelTaskCreator(user_wait_time=user_wait_time)

        for model in self.models.values():
            t = task_creator.create_task(model)
            if t is not None:
                # Record outstanding task info for stale-result detection
                model.outstanding_task_ID = t.task_id
                model.outstanding_task_kind = t.task_kind
                tasks.append(t)

                log.debug(
                    'Scheduled task: path=%s task_id=%s task_kind=%s class=%s '
                    '(gen_vgs=%s gen_decomps=%s)',
                    model.rel_path,
                    t.task_id,
                    t.task_kind.name if t.task_kind else None,
                    type(t).__name__,
                    getattr(t, 'gen_vgs', None),
                    getattr(t, 'gen_decomps', None),
                )

        if tasks:
            log.debug(
                'get_next_tasks summary: %s',
                [
                    {
                        'path': t.rel_path,
                        'task_kind': t.task_kind.name if t.task_kind else None,
                        'class': type(t).__name__,
                        'task_id': t.task_id,
                    }
                    for t in tasks
                ],
            )
        else:
            log.debug('get_next_tasks summary: no tasks scheduled')

        log.debug('Returning %d tasks', len(tasks))
        return tasks

    def abs_path_in_models(self, path: str) -> bool:
        """
        Return True if the specified path in the models
        """
        return self.make_path_relative(path) in self.models

    def make_path_relative(self, path: str) -> str:
        """
        Return True/False if the specified path is relative to the base source code directory
        """
        return str(Path(path).relative_to(self.src_dir_abs_path))

    def get_model_by_path(self, rel_path: str) -> Model | None:
        """
        Returns Model object for specified relative path, None if it's missing
        """

        if rel_path in self.models:
            return self.models[rel_path]
        return None

    def gather_deep_dependencies(self, rel_path: str) -> list[Model]:
        """
        Return list of all "deep" dependencies for the specified relative path
        """

        if rel_path not in self.models:
            return []

        def rec_helper(curr_model: Model):
            res = []
            for d in curr_model.dependencies:
                assert isinstance(d, Model)
                res.extend(rec_helper(d) + [d])
            return res

        curr_model = self.models[rel_path]
        return list(set(rec_helper(curr_model) + [curr_model]))

    def gen_consolidated_model(self, rel_path: str) -> tuple[str, dict[str, Model]]:
        """
        Generate a single consolidated model from the achnor model specified by (rel_path).
        """

        if rel_path not in self.models:
            raise Exception(f'Specified model {rel_path} is not a model!')

        # we'll gather the relevant models (subgraph) and then sort it
        relevant_models: list[Model] = self.gather_deep_dependencies(rel_path)
        sorted_list: list[Model] = ModelDepContext.topological_sort(relevant_models)

        # now we'll consolidate the code
        iml = ''
        for model in sorted_list:
            iml += (
                f'(* Starting {model.rel_path} *)\n'
                + (str(model.iml_code()) if model.iml_code() else '')
                + f'\n(* Ending {model.rel_path} *)\n'
            )

        return iml, {m.rel_path: m for m in relevant_models}

    def process_filesystem_event(self, event: FileSystemEvent):
        """
        Note that the provided paths are absolute!
        We receive events that concern both - Python and IML files.
        """

        # Let's figure out if we have a source code file or we have
        ext = Path(event.abs_path1).suffix
        if ext == '.py':
            self.process_fs_event_src_code(event)
        elif ext == '.iml':
            self.process_fs_event_iml_code(event)
        else:
            log.warning(
                f'MetaModel asked to process FS event for unknown extension: {ext}'
            )

    def process_fs_event_iml_code(self, event: FileSystemEvent):
        """
        Process FileSystem event when it's for a model
        """

        if event.action_type in [
            FileSystemEventType.CREATED,
            FileSystemEventType.MODIFIED,
        ]:
            # Let's first figure out what the Model path would be for this guy

            rel_path = self.make_path_relative(event.abs_path1)
            model_path = str(Path(rel_path).with_suffix('.py'))

            if model_path in self.models:
                try:
                    with open(event.abs_path1) as infile:
                        imlCode = infile.read()
                except Exception as e:
                    log.error(f'Failed to read IML file: {event.abs_path1}: {e}')
                    return

                self.models[model_path].set_iml_model(imlCode, record_time=True)
                self.models[model_path].expected_iml_on_disk = imlCode

                log.info(f'Updated IML code for {model_path}')
            else:
                log.info(
                    f'Received IML file for unknown model. Implied model path is: {model_path}'
                )
        else:
            log.info(
                f'We only support CREATED/MODIFIED FS events for IML models. Got: [{event.action_type}]'
            )

    def process_fs_event_src_code(self, event: FileSystemEvent):
        """
        Process FileSystem event when it's for a source code file
        """

        if event.action_type == FileSystemEventType.CREATED:
            if self.make_path_relative(event.abs_path1) in self.models:
                log.warning(
                    'Attempting to create a model with a path that already exists!'
                )
                return
            else:
                log.info(f'Creating a model for {event.abs_path1}')

                abs_path = event.abs_path1
                rel_path = self.make_path_relative(abs_path)
                try:
                    with open(abs_path) as infile:
                        srcCode = infile.read()
                except Exception:
                    log.error(f'Failed to read contents of the created file {abs_path}')
                    return

                self.models[rel_path] = Model(rel_path=rel_path, src_code=srcCode)
                log.info(f"We've now created a new model for path={rel_path}.")

        elif event.action_type == FileSystemEventType.MODIFIED:
            # A source file has been modified, so we will update our model info
            log.info(f'Processing modified source file: {event.abs_path1}')

            abs_path = event.abs_path1
            rel_path = self.make_path_relative(abs_path)

            try:
                with open(abs_path) as infile:
                    srcCode = infile.read()
            except Exception as e:
                log.error(
                    f'Failed to open a modified file: fullpath={abs_path}; error={str(e)}'
                )
                return

            if rel_path in self.models:
                # Let's now update the model's source code and record when we've done it
                self.models[rel_path].set_src_code(srcCode, True)
                log.info(
                    f"We've now updated source code for model with path=[{abs_path}]"
                )
            else:
                log.info(
                    'Modified file does not have a corresponding model. Creating it.'
                )
                self.models[rel_path] = Model(rel_path=rel_path, src_code=srcCode)
                log.info(f"We've now created a new model for path={abs_path}.")

        elif event.action_type == FileSystemEventType.DELETED:
            # We need to process a deleted file event
            log.info(f'Processing deleted source file: {event.abs_path1}')
            rel_path = self.make_path_relative(event.abs_path1)
            if rel_path in self.models:
                self.models[rel_path].set_src_code(None)
            else:
                log.error(
                    f'Attempting to process delete event for a non-existent model {event.abs_path1}'
                )

        elif event.action_type == FileSystemEventType.MOVED:
            # We need to process a deleted file event
            log.info(f'Processing a moved file: {event.abs_path1} to {event.abs_path2}')
            rel_path1 = self.make_path_relative(event.abs_path1)
            rel_path2 = self.make_path_relative(event.abs_path2)

            if rel_path1 in self.models and rel_path2 not in self.models:
                # self._models[rel_path].setSrcCode(None)
                self.models[rel_path2] = self.models[rel_path1]
                self.models[rel_path2].rel_path = rel_path2
                del self.models[rel_path1]
        else:
            log.error(
                "Something went wrong. We've gotten an unknown event type: {event.action_type}"
            )

    @staticmethod
    def _canonicalize_cycle_paths(rel_paths: list[str]) -> list[str]:
        # rel_paths is a closed walk: p[0] == p[-1]
        if len(rel_paths) <= 2:
            return rel_paths

        core = rel_paths[:-1]  # drop closing repeat
        # pick rotation that starts with smallest rel_path
        min_i = min(range(len(core)), key=lambda i: core[i])
        rotated = core[min_i:] + core[:min_i]
        return rotated + [rotated[0]]

    @staticmethod
    def _edge_paths_from_rel_paths(rel_paths: list[str]) -> list[tuple[str, str]]:
        return list(zip(rel_paths, rel_paths[1:]))

    def recompute_dep_cycles(self) -> None:
        """
        Recompute dependency cycles based on current live model graph.
        """
        models = sorted(self.models.values(), key=lambda m: m.rel_path)
        cycles = ModelDepContext.detect_cycles_in_models(models)

        summaries: list[CycleSummary] = []
        for c in cycles:
            rp = self._canonicalize_cycle_paths(c.rel_paths)
            ep = self._edge_paths_from_rel_paths(rp)
            summaries.append(CycleSummary(rel_paths=rp, edge_paths=ep))

        summaries.sort(key=lambda s: (len(s.rel_paths), s.rel_paths))
        dedup: dict[tuple[str, ...], CycleSummary] = {}
        for s in summaries:
            dedup[tuple(s.rel_paths)] = s

        self.dep_cycles = list(dedup.values())

    def run_file_sync(self):
        """
        Go through the source code directory and update the models.
        """

        log.info(f'Building FS from {self.src_dir_abs_path}')
        try:
            self.models = model_filesystem_sync(
                models=self.models,
                src_dir_abs_path=self.src_dir_abs_path,
                language=self.language,
            )
        except Exception as e:
            msg = f'Failed to perform model_filesync: {e}'
            log.error(msg)
            raise Exception(msg)

        self.recompute_dep_cycles()

        if not self.dep_cycles:
            log.info('FileSync ran successfully! No cycles detected')
        else:
            log.info(f'FileSync finished. Cycles detected: {self.dep_cycles}')

    def apply_sketch(self, sketch: Sketch) -> None:
        """
        Apply sketch back to the MetaModel
        """

        raise NotImplementedError('Sketch application is not implemented yet!')

    @staticmethod
    def fromJSON(j: str | dict):
        """
        Return an object from JSON file
        """
        if isinstance(j, str):
            return MetaModel.model_validate_json(
                j, context={'relink_from_str_edges': True}
            )
        else:
            return MetaModel.model_validate(j, context={'relink_from_str_edges': True})

    def toJSON(self):
        """
        Return a JSON object
        """
        return self.model_dump_json()

    def summary(self):
        """Return summary of the meta model state"""
        allFormStatuses = [
            FormalizationStatus.UNKNOWN,
            FormalizationStatus.TRANSPARENT,
            FormalizationStatus.INADMISSIBLE,
            FormalizationStatus.EXECUTABLE_WITH_APPROXIMATION,
            FormalizationStatus.ADMITTED_WITH_OPAQUENESS,
        ]

        frmStatuses = {s: 0 for s in allFormStatuses}
        for model in self.models.values():
            frmStatuses[model.formalization_status()] += 1

        def count(f):
            return sum(len(f(m)) for m in self.models.values())

        return {
            'frm_statuses': frmStatuses,
            **{
                label: count(f)
                for label, f in [
                    ('num_models', lambda _: [None]),
                    ('num_vgs', lambda m: m.verification_goals()),
                    ('num_opaque_funcs', lambda m: m.opaque_funcs()),
                    ('num_decomps', lambda m: m.decomps()),
                ]
            },
        }

    def eq_key(self) -> dict:
        return {
            'src_dir_abs_path': self.src_dir_abs_path,
            'language': self.language,
            'dep_cycles': [c.model_dump() for c in self.dep_cycles],
            'models': {
                rel: {
                    'rel_path': m.rel_path,
                    'src_code': m.src_code,
                    'src_language': m.src_language,
                    'context': m.context,
                    'iml_code_frozen': m.iml_code_frozen,
                    'outstanding_task_ID': m.outstanding_task_ID,
                    'outstanding_task_kind': (
                        m.outstanding_task_kind.name
                        if m.outstanding_task_kind
                        else None
                    ),
                    'analysis_vgs_attempted': m.analysis_vgs_attempted,
                    'analysis_decomps_attempted': m.analysis_decomps_attempted,
                    'str_dependencies': sorted(list(m.str_dependencies)),
                }
                for rel, m in sorted(self.models.items(), key=lambda kv: kv[0])
            },
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MetaModel):
            return NotImplemented
        return self.eq_key() == other.eq_key()



#
#   Imandra Inc.
#
#   codelogician/sketch/task.py
#

from enum import Enum

from imandra.u.agents.code_logician.base.region_decomp import RegionDecomp
from imandra.u.agents.code_logician.base.vg import VG
from pydantic import BaseModel


class SketchOperation(Enum):
    INIT = 0
    DELETE = 1
    INSERT = 2
    CHANGE = 3


class SketchChange(BaseModel):
    """
    Sketch change operation
    """

    change_type: SketchOperation = SketchOperation.CHANGE
    in_text: str = ''


class SketchChgSetModel(SketchChange, BaseModel):
    """
    Sketch Change Set Model operation (for initializations and for overrides)
    """

    new_iml_code: str


class SketchChgInsertDef(SketchChange, BaseModel):
    """
    Sketch Change Insert Definition operation
    """

    new_def_code: str


class SketchChgModifyDef(SketchChange, BaseModel):
    """
    Sketch Change Modify Definition operation
    """

    def_name_or_vg_idx: str | int
    new_def_body: str


class SketchChgDeleteDef(SketchChange, BaseModel):
    """
    Sketch Change Deletion Definition operation
    """

    def_name_or_vg_idx: str | int


class SketchChangeTask(BaseModel):
    """
    Sketch Change Task
    """

    sketch_id: str
    iml_code: str
    change: SketchChange


class SketchChangeResult(BaseModel):
    """
    Result of running ImandraX on the updated IML code
    """

    task: SketchChangeTask
    success: bool
    error: str | None = None
    vgs: list[VG]
    decomps: list[RegionDecomp]

    def toJSON(self):
        return self.model_dump_json()



#   
#   Imandra Bank
#
#   account.py
#

from .currency import Currency
from .transaction import DebitTransaction, CreditTransaction

class Account:
  """ Account class to represent a bank account """

  LIABILITY = 0
  ASSET = 1

  def __init__(self, customer: str, name: str, currency: int, account_type : int, account_number: str, starting_balance: float = 0.0):
    """ Initialize an account with account number and balance """

    self._account_type  = account_type      # type of account (liability or asset)    
    self._customer      = customer          # customer name
    self._name          = name              # account name
    self._number        = account_number    # account number 
    self._balance       = starting_balance  # current balance
    self._transactions  = []                # list of transactions

  def account_type (self):
    """ Get the type of the account """
    return self._account_type

  def customer(self):
    """ Get the customer associated with the account """
    return self._customer
  
  def name(self):
    """ Get the name of the account """
    return self._name

  def number(self):
    """ Get the account number """
    return self._number

  def currency(self):
    """ Get the currency of the account """
    return self._currency

  def balance(self):
    """ return the current balance of the account """
    return self._balance
  
  def transactions(self):
    """ Get the list of transactions in the account """
    return self._transactions

  def add_transaction(self, transaction):
    """ Add a transaction to the account """

    if isinstance(transaction, DebitTransaction):
      if self._type() ==  Account.ASSET and self._balance >= transaction.amount:
        self._balance -= transaction.amount
      else:
        raise ValueError("Insufficient balance for debit transaction")

    elif isinstance(transaction, CreditTransaction):
      if self._type() == Account.LIABILITY:
        self._balance += transaction.amount
      elif self._type() == Account.ASSET:
        self._balance -= transaction.amount
      else:
        raise ValueError("Invalid account type for credit transaction")
    
    else:
      raise ValueError("Invalid transaction type")

    self._transactions.append(transaction)

    def __repr__(self):
      """ String representation of the account """
      return f"Account(account_number={self.account_number}, balance={self.balance})"

if __name__ == "__main__":
    # Example usage
    account = Account("base_account", "123456789", 1000.0)



#
#   Imandra Inc.
#
#   cl_caller.py
#

import asyncio
from typing import Any

import dotenv
from imandra.u.agents import create_thread_sync, get_remote_graph
from imandra.u.agents.code_logician.command import (
    AgentFormalizerCommand,
    Command,
    GenFormalizationDataCommand,
    GenModelCommand,
    GenRegionDecompsCommand,
    GenVgsCommand,
    InitStateCommand,
)
from imandra.u.agents.code_logician.graph import GraphState

from codelogician.strategy.cl_agent_state import CLAgentState
from codelogician.util import get_imandra_uni_key, maybe

dotenv.load_dotenv('.env')


async def call_cl_agent(src_code: str, language: str, artifacts: bool = True):
    """
    src_code - source code we'll send to the agent
    language - programming language we should use
    artifacts - should we generate VGs/Decomps from the requests
    """

    graph = get_remote_graph('code_logician', api_key=get_imandra_uni_key())
    create_thread_sync(graph)

    gs = GraphState()

    #
    commands: list[Command] = [
        # Will initialize the state with specified Python source program
        InitStateCommand(src_code=src_code, src_lang=language),
        # Will gather relevant formalization data (required to create the model)
        AgentFormalizerCommand(
            no_gen_model_hitl=True,
            max_tries_wo_hitl=3,
            max_tries=3,
            no_check_formalization_hitl=True,
            no_refactor=False,
        ),
    ]

    # Create a list of commands for CL to execute
    if artifacts:
        more_commands = [GenVgsCommand(), GenRegionDecompsCommand()]
        commands.extend(more_commands)

    gs = gs.add_commands(commands)

    progress: dict[str, str] | None = None
    result: tuple[GraphState, dict[str, Any] | None] | None = None
    async for item in gs.stream(graph):
        if isinstance(item, dict):
            progress = item
            print(f'Currently running: {progress["step_name"]}')
        elif isinstance(item, tuple):
            result = item

    assert result is not None
    assert progress is not None

    gs: GraphState
    gs, _ = result

    return gs.last_fstate


def cl_autoformalize(src_code: str):
    """
    Return the object containing results of running CodeLogician agent
    """

    def mk_agent_state(f):
        return CLAgentState(
            status=f.status,
            src_code=f.src_code,
            iml_code=f.iml_code,
            vgs=f.vgs,
            region_decomps=f.region_decomps,
            opaque_funcs=f.opaque_funcs,
        )

    graph = get_remote_graph('code_logician', api_key=get_imandra_uni_key())
    create_thread_sync(graph)

    gs = GraphState()

    # Create a list of commands for CL to execute
    gs = gs.add_commands(
        [
            InitStateCommand(
                src_code=src_code, src_lang='python'
            ),  # Will initialize the state with specified Python source program
            GenFormalizationDataCommand(),  # Will gather relevant formalization data (required to create the model)
            GenModelCommand(),  # Will attempt to generate the formalized model
        ]
    )

    res = asyncio.run(gs.run(graph))  # Run the agent
    return maybe(mk_agent_state, res[0].last_fstate)



#
#   Imandra Bank
#
#   asset.py
#

from currency import Currency

class AssetType:
  """ Enum for asset types """
  CASH = 0
  STOCK = 1
  BOND = 2
  REAL_ESTATE = 3
  OTHER = 4

  def to_string(self):
    """ Convert the asset type to a string """
    return {
      self.CASH: "Cash",
      self.STOCK: "Stock",
      self.BOND: "Bond",
      self.REAL_ESTATE: "Real Estate",
      self.OTHER: "Other"
    }.get(self, "Unknown")

class Asset:
  """ Asset class to represent a financial asset """

  def __init__(self, name : str, asset_type : int, value: float):
    """ Initialize an asset with a value """
    if value < 0:
      raise ValueError("Asset value cannot be negative")
    self._value = value
    self._type = asset_type
    self._name = name

  def __repr__(self):
    return f"Asset(name={self._name}, value={self.value})"

  def __eq__(self, other):
    if not isinstance(other, Asset):
      return False
    return self.value == other.value

  def __hash__(self):
    return hash(self.value)

  def to_dict(self):
    """ Convert the asset to a dictionary """
    return {"value": self.value}

  @staticmethod
  def from_dict(data):
    """ Create an asset from a dictionary """
    return Asset(data['value'])

if __name__ == "__main__":
  # Example usage
  asset = Asset(1000.0)
  print(asset)

  # Convert to dictionary
  asset_dict = asset.to_dict()
  print(asset_dict)

  # Create from dictionary
  new_asset = Asset.from_dict(asset_dict)
  print(new_asset)



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import List


class Priority(Enum):
    LOW = 1
    MEDIUM = 2
    HIGH = 3


@dataclass
class User:
    id: int
    name: str


@dataclass
class Task:
    id: int
    title: str
    description: str
    priority: Priority
    deadline: datetime
    assigned_to: User | None = None
    completed: bool = False


@dataclass
class Project:
    id: int
    name: str
    tasks: List[Task] = field(default_factory=list)



#
# Imandra Inc.
#
# endpts_sketches.py
#

import asyncio
import logging

from fastapi import HTTPException

from codelogician.sketch.sketch import (
    SketchChange,
    SketchChangeResult,
    SketchChangeTask,
)
from codelogician.sketch.task import (
    SketchChgDeleteDef,
    SketchChgInsertDef,
    SketchChgModifyDef,
    SketchChgSetModel,
)
from codelogician.strategy.events import (
    SketchChangeEvent,
    SketchChangeResultEvent,
)
from codelogician.strategy.state import StrategyState
from codelogician.strategy.worker import run_sketch_task

from ..cl_server import CLServer

log = logging.getLogger(__name__)


def register_sketches_endpoints(app: CLServer):  # noqa: C901
    """Register sketch-related endpoints"""

    @app.get('/sketches/search', operation_id='search_sketches')
    async def search(query: str) -> list:
        """
        Search for specific sketch
        """
        return app.search_sketches(query or '')

    @app.get('/sketches/list', operation_id='list_sketches')
    def get_sketches_list() -> dict[str, list[dict[str, str]]]:
        """
        List all of the sketches available the currently selected strategy.
        Returns a list of dictionaries, references by strategy id and mapped to a list of sketch IDs.

        'sketch_id': sketch_id,
        'anchor_model': self.sketches[sketch_id].anchor_model_path

        Example: [
            {'sketch_id': "SketchID1", 'anchormodel': "path1.py"},
            {'sketch_id': "SketchID2", 'anchromodel': "path2.py"}
        """

        return app.list_sketches()

    @app.post('/sketches/create', operation_id='create_sketch')
    def create_sketch(anchor_model_path: str):
        """
        Create a new sketch for this strategy
        """

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            strat_state: StrategyState = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        if anchor_model_path not in strat_state.curr_meta_model.models:
            raise HTTPException(
                status_code=404, detail=f'Unknown model path: {anchor_model_path}'
            )

        try:
            sketch_id = app.create_sketch(strategy_id, anchor_model_path)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to create a new sketch:{e}'
            )

        return sketch_id

    @app.post('/sketches/{sketch_id}/try_change', operation_id='try_sketch_change')
    async def try_sketch_change(
        sketch_id: str, change: SketchChange, commit_on_success: bool = True
    ) -> SketchChangeResult | str:
        """
        Try a sketch change (optionally and if successful, apply it - by default, this is True)
        """

        sketch = app.get_sketch_from_sketch_id(sketch_id)

        if sketch is None:
            raise HTTPException(
                status_code=404, detail=f'Could not locate sketch with ID={sketch_id}'
            )

        try:
            result_iml = sketch.process_change(sketch_id, change)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f'Could not apply change: {e}')

        try:
            change_result = asyncio.run(
                run_sketch_task(
                    SketchChangeTask(
                        sketch_id=sketch_id, change=change, iml_code=result_iml
                    )
                )
            )
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Error during call to ImandraX: {e}'
            )

        if (
            commit_on_success
            and change_result is not None
            and change_result.error is None
        ):
            strat = app.get_strat_for_sketch_id(sketch_id=sketch_id)
            if strat:
                strat.add_event(
                    SketchChangeResultEvent(
                        sketch_id=sketch_id, change_result=change_result
                    )
                )

                log.info(
                    f'Added SketchChangeResult to strategy[id={strat.state().strat_id}]'
                )
            else:
                log.error(f'Failed to find strategy for sketch_id = {sketch_id}')

        if change_result is None:
            return 'N//A'
        else:
            return change_result

    @app.post('/sketches/{sketch_id}/change', operation_id='apply_sketch_change')
    def apply_sketch_change(
        sketch_id: str,
        change: SketchChgSetModel
        | SketchChgInsertDef
        | SketchChgModifyDef
        | SketchChgDeleteDef,
    ):
        """
        Apply sketch change
        """

        if sketch_id not in app.all_sketch_ids():
            raise HTTPException(
                status_code=404, detail=f'Sketch with id={sketch_id} was not found!'
            )

        strat = app.get_strat_for_sketch_id(sketch_id)

        if strat is None:
            raise HTTPException(
                status_code=400,
                detail=f'Could not locate strategy for sketch={sketch_id}',
            )

        strat.add_event(SketchChangeEvent(sketch_id=sketch_id, change=change))

        return 'OK'

    @app.post('/sketches/{sketch_id}/rollback', operation_id='rollback_changes')
    def rollback_changes(sketch_id: str, target_state_id: int | None = None):
        """
        Rollback changes. If `target_state_id` is specified, then the Sketch will go back to the specified state ID if possible.
        If `target_state_id` is not specified, then the Sketch will rollback the last change, unless it's the initial one.
        """

        sketch = app.get_sketch_from_sketch_id(sketch_id)

        if sketch is None:
            raise HTTPException(
                status_code=404, detail=f'Sketch with id={sketch_id} was not found!'
            )

        try:
            sketch.rollback(target_state_id=target_state_id)
        except Exception as e:
            raise HTTPException(
                status_code=403, detail=f'Failed to rollback the state: {e}'
            )

        return 'OK'

    @app.post('/sketches/{sketch_id}/state', operation_id='')
    def get_latest_sketch_state(sketch_id: str):
        """
        Return the latest sketch state
        """

        if sketch_id not in app.all_sketch_ids():
            raise HTTPException(
                status_code=404, detail=f'Could not locate sketch with id = {sketch_id}'
            )

        return 'OK'



#
#   Imandra Inc.
#
#   codelogician/commands/cmd_server.py
#

from __future__ import annotations

from typing import Annotated

import typer

from codelogician.util import setup_logging

from .server.cmd_metamodel import app as metamodel_app
from .server.cmd_model import app as model_app
from .server.cmd_sketches import app as sketches_app
from .server.cmd_strategy import app as strategy_app

app = typer.Typer(name='server', no_args_is_help=True)


@app.callback()
def _root(
    ctx: typer.Context,
    base_url: str = typer.Option(
        'http://localhost:8000',
        '--base-url',
        envvar='CODELOGICIAN_SERVER_URL',
        help='Base URL for the CodeLogician server.',
    ),
    timeout: float = typer.Option(30.0, '--timeout', help='HTTP timeout in seconds.'),
    json_out: bool = typer.Option(
        True, '--json/--no-json', help='Pretty-print JSON output.'
    ),
):
    ctx.obj = {'base_url': base_url, 'timeout': timeout, 'json_out': json_out}


app.add_typer(metamodel_app, name='metamodel')
app.add_typer(model_app, name='model')
app.add_typer(sketches_app, name='sketches')
app.add_typer(strategy_app, name='strategy')


# fmt: off
@app.command(name='start', help='Start the server')
def run_server_cmd(
    dir   : Annotated[str, typer.Argument(help='Starting directory')],
    state : Annotated[str | None, typer.Option(help='Server state path')] = None,
    clean : Annotated[bool, typer.Option(help='Disregard existing strategy caches')] = False,
    config: Annotated[str, typer.Option(help='Configuration path')] = 'config/server_config.yaml',
    debug : Annotated[bool, typer.Option(help='Debug mode')] = False,
    addr  : Annotated[str, typer.Option(help='Server host/port')] = 'http://127.0.0.1:8000',
):
    from codelogician.server.main import run_server

    setup_logging(debug=debug)
    run_server(dir=dir, state=state, clean=clean, config=config, addr=addr)
# fmt: on


if __name__ == '__main__':
    app()



"""Core mathematical operations."""

from .arithmetic import add, divide, multiply, subtract
from .comparison import absolute_value, maximum, minimum

__all__ = [
    "absolute_value",
    "add",
    "divide",
    "maximum",
    "minimum",
    "multiply",
    "subtract",
]



from pydantic import BaseModel, Field


class Session(BaseModel):
    """A session identifier"""

    id: str = Field(description="The session's unique ID (e.g a uuid)")


class SessionCreate(BaseModel):
    """Create a new session"""

    po_check: bool = Field(default=True, description='Do we check Proof Obligations?')
    api_version: str = Field(description='the API types version (mandatory)')


class SessionOpen(BaseModel):
    """Reconnect to the given session"""

    id: Session | None = Field(
        default=None, description="The session's unique ID (e.g a uuid)"
    )
    api_version: str = Field(description='the API types version (mandatory)')



#
#   Imandra Inc.
#
#   help.py
#

from functools import partial
from pathlib import Path

from rich.style import Style
from rich.text import Text
from textual.containers import Horizontal, VerticalScroll
from textual.screen import Screen
from textual.widgets import Footer, Markdown, Tree

from codelogician.doc.utils.docs import walk_directory
from codelogician.util import maybe, snd

from ..common import MyHeader

TOGGLE_STYLE = Style.from_meta({'toggle': True})


class TOC(Tree):
    def __init__(self, nodes):
        Tree.__init__(self, 'CodeLogician Help')

        def add_children(node, sub_nodes):
            for n in sub_nodes:
                add_subtree(node, n)

        def add_subtree(dest, src_node):
            kind, data = src_node
            if kind == 'leaf':
                _, doc = data
                dest.add_leaf(snd(data).title, data=data)
            else:
                data, sub_nodes = data
                add_children(dest.add(snd(data).title, data=data), sub_nodes)

        self.root.expand()
        add_children(self.root, nodes)

    def render_label(self, node, base_style, style):
        node_label = node._label.copy()
        node_label.stylize(style)

        def toggled_icon():
            icon = self.ICON_NODE_EXPANDED if node.is_expanded else self.ICON_NODE
            return icon, base_style + TOGGLE_STYLE

        prefix = toggled_icon() if node._allow_expand else ('', base_style)
        return Text.assemble(prefix, node_label)


def code_block(lang, code):
    return f'```{lang}\n{code}\n```\n'


def fix_code_blocks(s):
    return s.replace('```iml', '```ocaml')


content_wrappers = {
    '': fix_code_blocks,
    '.md': fix_code_blocks,
    '.iml': partial(code_block, 'ocaml'),
    '.py': partial(code_block, 'python'),
    '.ts': partial(code_block, 'typescript'),
}


class HelpScreen(Screen):
    DEFAULT_CSS = 'TOC { width: 25% } VerticalScroll { width: 75% }'

    def __init__(self):
        Screen.__init__(self)
        self.title = 'Help'

    def compose(self):
        doc_path = Path(__file__).parent / '../../doc/data'
        nodes = walk_directory(doc_path.resolve())

        yield MyHeader()
        with Horizontal():
            yield TOC(nodes)
            with VerticalScroll():
                markdown = Markdown(
                    '# CodeLogician\nClick the tree to read the docs.', id='content'
                )
                # markdown.code_indent_guides = False
                yield markdown
        yield Footer()

    def on_tree_node_selected(self, event):
        def select_node(node_data):
            path, doc = node_data
            content = content_wrappers[Path(path).suffix](doc.content)
            text = f'# {doc.order}: {doc.title}\n{content}'
            self.query_one('#content', Markdown).update(text)

        maybe(select_node, event.node.data)
